"""Views module."""
from typing import Any, Dict, List, Tuple, Union, cast

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.forms import Form, ModelForm
from django.urls import path, reverse_lazy
from django.utils.translation import ugettext as _
from django.views.generic.base import View
from django.views.generic.detail import DetailView
from django.views.generic.edit import CreateView, DeleteView, UpdateView
from django.views.generic.list import ListView

from braces.views import FormMessagesMixin, LoginRequiredMixin, PermissionRequiredMixin
from django_filters import FilterSet
from django_tables2 import SingleTableView, Table
from django_tables2.export.views import ExportMixin

from vega_admin.forms import ListViewSearchForm
from vega_admin.mixins import (
    CRUDURLsMixin,
    DeleteViewMixin,
    DetailViewMixin,
    ListViewSearchMixin,
    ObjectTitleMixin,
    ObjectURLPatternMixin,
    PageTitleMixin,
    SimpleURLPatternMixin,
    VegaFormMixin,
    VegaOrderedQuerysetMixin,
    VerboseNameMixin,
)
from vega_admin.utils import (
    customize_modelform,
    get_filterclass,
    get_listview_form,
    get_modelform,
    get_table,
)


# pylint: disable=too-many-ancestors,bad-continuation
class VegaListView(
    VerboseNameMixin,
    ListViewSearchMixin,
    PageTitleMixin,
    CRUDURLsMixin,
    ExportMixin,
    SingleTableView,
    SimpleURLPatternMixin,
    VegaOrderedQuerysetMixin,
    ListView,
):
    """vega-admin Generic List View."""

    template_name = f"vega_admin/{settings.VEGA_TEMPLATE}/list.html"


class VegaCreateView(
    FormMessagesMixin,
    PageTitleMixin,
    VerboseNameMixin,
    VegaFormMixin,
    CRUDURLsMixin,
    SimpleURLPatternMixin,
    CreateView,
):
    """vega-admin Generic Create View."""

    template_name = f"vega_admin/{settings.VEGA_TEMPLATE}/create.html"
    form_valid_message = _(settings.VEGA_FORM_VALID_CREATE_TXT)
    form_invalid_message = _(settings.VEGA_FORM_INVALID_TXT)


class VegaDetailView(
    PageTitleMixin,
    VerboseNameMixin,
    CRUDURLsMixin,
    ObjectURLPatternMixin,
    ObjectTitleMixin,
    DetailViewMixin,
    DetailView,
):
    """vega-admin Generic Detail View."""

    template_name = f"vega_admin/{settings.VEGA_TEMPLATE}/read.html"


class VegaUpdateView(
    FormMessagesMixin,
    PageTitleMixin,
    VerboseNameMixin,
    VegaFormMixin,
    CRUDURLsMixin,
    ObjectURLPatternMixin,
    ObjectTitleMixin,
    UpdateView,
):
    """vega-admin Generic Update View."""

    template_name = f"vega_admin/{settings.VEGA_TEMPLATE}/update.html"
    form_valid_message = _(settings.VEGA_FORM_VALID_UPDATE_TXT)
    form_invalid_message = _(settings.VEGA_FORM_INVALID_TXT)


class VegaDeleteView(
    FormMessagesMixin,
    PageTitleMixin,
    VerboseNameMixin,
    DeleteViewMixin,
    CRUDURLsMixin,
    ObjectURLPatternMixin,
    ObjectTitleMixin,
    DeleteView,
):
    """vega-admin Generic Delete View."""

    template_name = f"vega_admin/{settings.VEGA_TEMPLATE}/delete.html"
    form_valid_message = _(settings.VEGA_FORM_VALID_DELETE_TXT)
    form_invalid_message = _(settings.VEGA_FORM_INVALID_TXT)


class VegaCRUDView:  # pylint: disable=too-many-public-methods
    """
    Creates generic CRUD views for a model automagically.

    The intention is for you to give it at least a model and get back an entire
    set of CRUD views, with everything you need to start doing CRUD actions
    straight away.
    """

    actions: List[str] = settings.VEGA_DEFAULT_ACTIONS
    protected_actions: Union[None, List[str]] = actions  # actions that require login
    permissions_actions: Union[None, List[str]] = actions
    view_classes: Dict[str, View] = {}
    list_fields: Union[None, List[str]] = None
    read_fields: Union[None, List[str]] = None
    search_fields: Union[None, List[str]] = None
    filter_fields: Union[None, List[str]] = None
    filter_class: Union[None, FilterSet] = None
    search_form_class: Union[None, Form, ModelForm] = ListViewSearchForm
    form_fields: Union[None, List[str]] = None
    create_fields: Union[None, List[str]] = None
    update_fields: Union[None, List[str]] = None
    table_attrs: Union[None, Dict[str, str]] = None
    table_actions: List[str] = [
        settings.VEGA_READ_ACTION,
        settings.VEGA_UPDATE_ACTION,
        settings.VEGA_DELETE_ACTION,
    ]
    form_class: Union[None, Form, ModelForm] = None
    create_form_class: Union[None, Form, ModelForm] = None
    update_form_class: Union[None, Form, ModelForm] = None
    table_class: Union[None, Table] = None
    paginate_by = 25
    crud_path: Union[None, str] = None
    order_by: Union[None, List[str], str] = None

    def __init__(self, model=None):
        """Initialize!."""
        if model is not None:
            self.model = model
        self.model_name = self.model._meta.model_name
        self.app_label = self.model._meta.app_label

        if self.crud_path is None:
            self.crud_path = self.model._meta.label_lower

    def get_actions(self):
        """Get actions."""
        custom_actions = self.get_view_classes().keys()
        custom_actions = [_ for _ in custom_actions if _ not in self.actions]
        return self.actions + custom_actions

    def get_view_classes(self):  # pylint: disable=no-self-use
        """Return the available views."""
        return self.view_classes

    def get_protected_actions(self):
        """Get list of actions that have login protection."""
        if isinstance(self.protected_actions, list):
            return self.protected_actions
        return []

    def get_permissions_actions(self):
        """Get list of actions that have permissions protection."""
        if isinstance(self.permissions_actions, list):
            return self.permissions_actions
        return []

    def get_permissions(self):
        """Get list of permission names associated with this CRUD view."""
        actions = self.get_actions()
        return [self.get_permission_for_action(action) for action in actions]

    def get_search_fields(self):
        """Get search fields for list view."""
        return self.search_fields

    def get_read_fields(self):
        """Get read fields for read view."""
        return self.read_fields

    def get_filter_fields(self):
        """Get filter fields for list view."""
        return self.filter_fields

    def get_search_form_class(self):
        """Get search form for list view."""
        if self.search_form_class:  # pylint: disable=using-constant-test
            return self.search_form_class

        # if there are filters then we generate the form
        filter_class = self.get_filter_class()
        if filter_class:
            return get_listview_form(
                model=self.model,
                fields=filter_class.Meta.fields,
                include_search=self.get_search_fields() is not None,
            )

        return None

    def get_createform_fields(self):
        """Get fields for create form."""
        if self.create_fields:
            return self.create_fields
        if self.form_fields:
            return self.form_fields

        return None

    def get_createform_class(self):
        """Get form class for create view."""
        chosen_form = None
        if self.create_form_class:
            chosen_form = self.create_form_class
        elif self.form_class:
            chosen_form = self.form_class

        if chosen_form is not None:
            return customize_modelform(cast(Union[Form, ModelForm], chosen_form))

        return get_modelform(model=self.model, fields=self.get_createform_fields())

    def get_updateform_fields(self):
        """Get fields for update form."""
        if self.update_fields:
            return self.update_fields
        if self.form_fields:
            return self.form_fields

        return None

    def get_updateform_class(self):
        """Get form class for create view."""
        chosen_form = None
        if self.update_form_class:
            chosen_form = self.update_form_class
        elif self.form_class:
            chosen_form = self.form_class

        if chosen_form is not None:
            return customize_modelform(cast(Union[Form, ModelForm], chosen_form))

        return get_modelform(model=self.model, fields=self.get_updateform_fields())

    def get_list_fields(self):
        """Get the list_fields."""
        return self.list_fields

    def get_table_actions(self):
        """Get the table actions."""
        return self.table_actions

    def get_table_attrs(self):
        """Get the table_attrs."""
        return self.table_attrs

    def get_table_class(self):
        """Get the table class."""
        if self.table_class:
            return self.table_class

        tables_kwargs = {"model": self.model}
        if isinstance(self.get_list_fields(), list):
            tables_kwargs["fields"] = self.get_list_fields()
        if isinstance(self.get_table_actions(), list):
            table_actions = [
                _ for _ in self.get_table_actions() if _ in self.get_actions()
            ]
            tables_kwargs["actions"] = self.get_action_urlnames(actions=table_actions)
        if isinstance(self.get_table_attrs(), dict):
            tables_kwargs["attrs"] = self.get_table_attrs()

        return get_table(**tables_kwargs)

    def get_filter_class(self):
        """Get the filter class."""
        if self.filter_class:
            return self.filter_class

        if self.filter_fields:
            filter_kwargs = {"model": self.model, "fields": self.get_filter_fields()}

            return get_filterclass(**filter_kwargs)

        return None

    def get_create_view_class(self):  # pylint: disable=no-self-use
        """Get view class for create action."""
        return VegaCreateView

    def get_update_view_class(self):  # pylint: disable=no-self-use
        """Get view class for update action."""
        return VegaUpdateView

    def get_read_view_class(self):  # pylint: disable=no-self-use
        """Get view class for read action."""
        return VegaDetailView

    def get_list_view_class(self):  # pylint: disable=no-self-use
        """Get view class for list action."""
        return VegaListView

    def get_delete_view_class(self):  # pylint: disable=no-self-use
        """Get view class for delete action."""
        return VegaDeleteView

    def get_success_url(self):  # pylint: disable=no-self-use
        """Get success_url."""
        return reverse_lazy(self.get_url_name_for_action(settings.VEGA_LIST_ACTION))

    def get_cancel_url(self):  # pylint: disable=no-self-use
        """Get cancel_url."""
        return self.get_success_url()

    # pylint: disable=no-self-use
    def enforce_permission_protection(self, view_class: View, action: str):
        """Ensure view class has permission protection."""
        has_perms_mixin = issubclass(view_class, PermissionRequiredMixin)
        has_login_mixin = issubclass(view_class, LoginRequiredMixin)

        options = {"permission_required": self.get_permission_for_action(action)}

        if not has_login_mixin and not has_perms_mixin:
            # add LoginRequiredMixin and PermissionRequiredMixin
            return type(
                f"{view_class.__name__}{settings.VEGA_PROTECTED_LABEL}",
                (LoginRequiredMixin, PermissionRequiredMixin, view_class),
                options,
            )

        if not has_login_mixin:
            # add LoginRequiredMixin
            return type(
                f"{view_class.__name__}{settings.VEGA_PROTECTED_LABEL}",
                (LoginRequiredMixin, view_class),
                options,
            )

        if not has_perms_mixin:
            # in this case the LoginRequiredMixin was set but not
            # PermissionRequiredMixin
            raise ImproperlyConfigured(
                _(
                    f"{settings.VEGA_PERMREQUIRED_NOT_SET_TXT} {view_class.__name__}"
                )  # noqa
            )

        return view_class

    # pylint: disable=no-self-use
    def enforce_login_protection(self, view_class: View):
        """Ensure view class has login protection."""
        if issubclass(view_class, LoginRequiredMixin):
            return view_class

        # add LoginRequiredMixin
        return type(
            f"{view_class.__name__}{settings.VEGA_PROTECTED_LABEL}",
            (LoginRequiredMixin, view_class),
            {},
        )

    def get_default_action_view_classes(self, action: str):
        """Get view class for default actions."""
        if action == settings.VEGA_LIST_ACTION:
            return self.get_list_view_class()
        if action == settings.VEGA_CREATE_ACTION:
            return self.get_create_view_class()
        if action == settings.VEGA_READ_ACTION:
            return self.get_read_view_class()
        if action == settings.VEGA_UPDATE_ACTION:
            return self.get_update_view_class()
        if action == settings.VEGA_DELETE_ACTION:
            return self.get_delete_view_class()

        # this action is set as a default action but has no defined view class
        raise Exception(settings.VEGA_INVALID_ACTION)

    def get_view_class_for_action(  # pylint: disable=too-many-branches
        self, action: str
    ):
        """Get the view for an action."""
        view_classes = self.get_view_classes()
        try:
            # return the view class if found
            view_class = view_classes[action]
        except KeyError:
            if action not in settings.VEGA_DEFAULT_ACTIONS:
                # this action is not supported
                raise Exception(settings.VEGA_INVALID_ACTION)

            # lets get the view class for the default actions
            view_class = self.get_default_action_view_classes(action)
        else:
            if action in self.get_permissions_actions():
                return self.enforce_permission_protection(view_class, action)

            if action in self.get_protected_actions():
                return self.enforce_login_protection(view_class)

            return view_class

        # lets go on and create the view class(es)
        options = {"model": self.model}
        # add some common useful CRUD urls
        if settings.VEGA_LIST_ACTION in self.get_actions():
            options["list_url"] = reverse_lazy(
                self.get_url_name_for_action(settings.VEGA_LIST_ACTION)
            )
            options["order_by"] = self.order_by
        if settings.VEGA_CREATE_ACTION in self.get_actions():
            options["create_url"] = reverse_lazy(
                self.get_url_name_for_action(settings.VEGA_CREATE_ACTION)
            )
        options["cancel_url"] = self.get_cancel_url()

        # add the success url
        if action in [
            settings.VEGA_CREATE_ACTION,
            settings.VEGA_UPDATE_ACTION,
            settings.VEGA_DELETE_ACTION,
        ]:
            options["success_url"] = self.get_success_url()

        # add the create form class
        if action == settings.VEGA_CREATE_ACTION:
            options["form_class"] = self.get_createform_class()

        # add the update form class and update url
        if action == settings.VEGA_UPDATE_ACTION:
            options["form_class"] = self.get_updateform_class()
            options["update_url_name"] = self.get_url_name_for_action(
                settings.VEGA_UPDATE_ACTION
            )

        # add the read url
        if action == settings.VEGA_READ_ACTION:
            options["read_url_name"] = self.get_url_name_for_action(
                settings.VEGA_READ_ACTION
            )
            options["fields"] = self.get_read_fields()

        # add the delete url
        if action == settings.VEGA_DELETE_ACTION:
            options["delete_url_name"] = self.get_url_name_for_action(
                settings.VEGA_DELETE_ACTION
            )

        # add the table class
        if action == settings.VEGA_LIST_ACTION:
            options["table_class"] = self.get_table_class()
            options["search_fields"] = self.get_search_fields()
            options["form_class"] = self.get_search_form_class()
            options["paginate_by"] = self.paginate_by
            options["filter_class"] = self.get_filter_class()

        inherited_classes: Tuple[Any, ...] = (view_class,)

        # permissions and login protection
        if action in self.get_permissions_actions():
            inherited_classes = (
                LoginRequiredMixin,
                PermissionRequiredMixin,
                view_class,
            )
            options["permission_required"] = self.get_permission_for_action(action)
        elif action in self.get_protected_actions():
            inherited_classes = (LoginRequiredMixin, view_class)

        # create and return the View class
        view_label = settings.VEGA_VIEW_LABEL
        return type(
            f"{self.model_name.title()}{action.title()}{view_label}",
            inherited_classes,  # the classes that we should inherit
            options,
        )

    # pylint: disable=no-self-use
    def get_url_name_for_action(self, action: str):
        """Return the url name for the action."""
        return f"{self.crud_path}-{action}"

    def get_permission_for_action(self, action: str):
        """Get permission for action."""
        return f"{self.app_label}.{action}_{self.model_name}"

    def get_action_urlname(self, action: str):
        """Get tuple of action and url name."""
        return (action, self.get_url_name_for_action(action))

    def get_action_urlnames(self, actions: list = None):
        """Get list of tuples of (action, url_name)."""
        if actions is None:
            actions = self.get_actions()
        return [self.get_action_urlname(_) for _ in actions]

    def get_url_pattern_for_action(self, view_class, action: str):
        """Return the url pattern for the provided action."""
        try:
            return view_class.derive_url_pattern(self.crud_path, action)
        except NameError:
            # the view does not know how to derive a url pattern
            # we are forced to try something and hope it works :(
            return f"{view_class.get_crud_path_pattern(self.crud_path, action)}/"

    def url_patterns(self, actions: list = None):
        """Return the URL patters for the selected actions in this CRUD view."""
        if actions is None:
            actions = self.get_actions()
        urls = []
        for action in actions:
            view_class = self.get_view_class_for_action(action=action)
            pattern = self.get_url_pattern_for_action(view_class, action)
            url_name = self.get_url_name_for_action(action=action)
            urls.append(path(pattern, view_class.as_view(), name=url_name))

        return urls



"""
vega-admin forms module
"""
from django import forms
from django.conf import settings
from django.utils.translation import ugettext as _

from crispy_forms.bootstrap import Field, FieldWithButtons
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Submit

from vega_admin.settings import VEGA_LISTVIEW_SEARCH_QUERY_TXT


class ListViewSearchForm(forms.Form):
    """
    search form for use in the vega-admin list view
    """

    q = forms.CharField(
        label=_(
            getattr(settings, 'VEGA_LISTVIEW_SEARCH_QUERY_TXT',
                    VEGA_LISTVIEW_SEARCH_QUERY_TXT)),
        required=False)

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.form_method = 'GET'
        self.helper.render_required_fields = True
        self.helper.form_show_labels = False
        self.helper.html5_required = True
        self.helper.form_id = 'vega-search-form'
        self.helper.form_class = 'form-inline'
        self.helper.field_template = \
            f'{settings.VEGA_CRISPY_TEMPLATE_PACK}/layout/inline_field.html'
        self.helper.layout = Layout(
            FieldWithButtons(
                Field('q', css_class="input-sm"),
                Submit(
                    'submitBtn',
                    _(settings.VEGA_LISTVIEW_SEARCH_TXT),
                    css_class='btn-sm'
                ),
            )
        )



"""
Module for vega-admin test forms
"""
from django import forms

from crispy_forms.bootstrap import Field, FormActions
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Submit

from vega_admin.forms import ListViewSearchForm

from .models import Artist, Song


class ArtistForm(forms.ModelForm):
    """
    Artist ModelForm class
    """

    class Meta:
        model = Artist
        fields = ["name"]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request", None)
        self.vega_extra_kwargs = kwargs.pop("vega_extra_kwargs", dict())
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.include_media = False
        self.helper.form_id = "artist"
        self.helper.layout = Layout(
            Field("name"),
            FormActions(
                Submit("submitBtn", "Submit", css_class="btn-success btn-block")
            ),
        )


class UpdateArtistForm(forms.ModelForm):
    """
    Artist ModelForm class with nothing extra
    """

    class Meta:
        model = Artist
        fields = ["name"]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request", None)
        self.vega_extra_kwargs = kwargs.pop("vega_extra_kwargs", dict())
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.include_media = False
        self.helper.form_id = "artist-update"
        self.helper.layout = Layout(
            Field("name"),
            FormActions(
                Submit("submitBtn", "Submit", css_class="btn-success btn-block")
            ),
        )


class PlainArtistForm(forms.ModelForm):
    """
    Artist ModelForm class with nothing extra
    """

    class Meta:
        model = Artist
        fields = ["name"]


class SongForm(forms.ModelForm):
    """
    Artist ModelForm class
    """

    class Meta:
        model = Song
        fields = ["name", "artist"]

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request", None)
        self.vega_extra_kwargs = kwargs.pop("vega_extra_kwargs", dict())
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_tag = True
        self.helper.render_required_fields = True
        self.helper.form_show_labels = True
        self.helper.html5_required = True
        self.helper.include_media = False
        self.helper.form_id = "song"
        self.helper.layout = Layout(
            Field("name"),
            Field("artist"),
            FormActions(
                Submit("submitBtn", "Submit", css_class="btn-success btn-block")
            ),
        )


class CustomSearchForm(ListViewSearchForm):
    """Custom search form"""



"""Views module for Vega Admin users app"""
from django.conf import settings
from django.contrib.auth.models import Group, User
from django.urls import reverse_lazy

from vega_admin.contrib.users.forms import (AddUserForm, EditUserForm,
                                            PasswordChangeForm)
from vega_admin.views import VegaCRUDView, VegaUpdateView


class ChangePassword(VegaUpdateView):  # pylint: disable=too-many-ancestors
    """
    Change Password view
    """

    form_class = PasswordChangeForm
    model = User

    def get_success_url(self):  # pylint: disable=no-self-use
        """Get success_url"""
        return reverse_lazy("auth.user-list")


class UserCRUD(VegaCRUDView):
    """
    CRUD view for users
    """

    model = User
    list_fields = ["id", "username", "email", "first_name", "last_name"]
    create_form_class = AddUserForm
    update_form_class = EditUserForm
    protected_actions = None
    permissions_actions = None
    view_classes = {settings.VEGA_CHANGE_PASSWORD_LABEL: ChangePassword}
    table_actions = [
        settings.VEGA_READ_ACTION,
        settings.VEGA_UPDATE_ACTION,
        settings.VEGA_CHANGE_PASSWORD_LABEL,
        settings.VEGA_DELETE_ACTION,
    ]
    order_by = ["-last_login", "first_name"]


class GroupCRUD(VegaCRUDView):
    """
    CRUD view for Groups
    """

    model = Group
    protected_actions = None
    permissions_actions = None
    list_fields = ["id", "name"]



"""Widgets module for django-vega-admin"""
from django.forms import DateInput, DateTimeInput, TimeInput


class VegaDateWidget(DateInput):
    """HTML5 Date input widget class"""

    input_type = "date"


class VegaTimeWidget(TimeInput):
    """HTML5 Time input widget class"""

    input_type = "time"


class VegaDateTimeWidget(DateTimeInput):
    """HTML5 Time input widget class"""

    input_type = "datetime-local"



"""module for crispy form utils."""
from typing import List, Optional

from django.conf import settings
from django.utils.translation import ugettext as _

from crispy_forms.bootstrap import FormActions
from crispy_forms.helper import FormHelper
from crispy_forms.layout import HTML, Div, Layout, Submit


def get_form_actions(  # pylint: disable=bad-continuation
    cancel_url: Optional[str] = None,
    submit_text: str = settings.VEGA_SUBMIT_TEXT,
    cancel_text: str = settings.VEGA_CANCEL_TEXT,
    button_div_css_class: str = "col-md-6",
) -> FormActions:
    """
    Return the FormActions class.

    :param cancel_url: the cancel url
    :param submit_text: the text for the submit button
    :param cancel_text: the text for the cancel button

    :return: form actions object

    """
    button_div = Div(css_class="col-md-12")
    if cancel_url:
        button_div.append(
            Div(
                HTML(
                    f"""<a href="{cancel_url}"
                            class="btn btn-default btn-block vega-cancel">
                            {_(cancel_text)}
                        </a>"""
                ),
                css_class=button_div_css_class,
            )
        )
    button_div.append(
        Div(
            Submit("submit", _(submit_text), css_class="btn-block vega-submit"),
            css_class=button_div_css_class,
        )
    )

    return FormActions(Div(button_div, css_class="row"))


def get_form_helper_class(  # pylint: disable=too-many-arguments,bad-continuation
    form_tag: bool = True,
    form_method: str = "POST",
    render_required_fields: bool = True,
    form_show_labels: bool = True,
    html5_required: bool = True,
    include_media: bool = True,
) -> FormHelper:
    """
    Return the base form helper class.

    :param form_tag: include form tag?
    :param form_method: form method
    :param render_required_fields: render required fields?
    :param form_show_labels: show form labels?
    :param html5_required: HTML5 required?
    :param include_media: include form media?

    :return: form helper class
    """
    helper = FormHelper()
    helper.form_tag = form_tag
    helper.form_method = form_method
    helper.render_required_fields = render_required_fields
    helper.form_show_labels = form_show_labels
    helper.html5_required = html5_required
    helper.include_media = include_media

    return helper


def get_layout(  # pylint: disable=bad-continuation
    formfields: List[str], with_actions: bool = False, cancel_url: str = "/"
) -> Layout:
    """Get layout class for crispy form helper.

    Arguments:
        formfields {List[str]} -- list of form fields

    Keyword Arguments:
        with_actions {bool} -- whether to include form actions (default: {False})
        cancel_url {str} -- the cancel url (default: {"/"})

    Returns:
        Layout -- the crispy forms Layout object

    """
    layout = Layout(*formfields)
    if with_actions is True:
        form_actions_class = get_form_actions(cancel_url=cancel_url)
        layout.append(form_actions_class)
    return layout



"""vega-admin mixins module."""
from typing import List

from django.conf import settings
from django.contrib import messages
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.db.models import ProtectedError, Q
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.text import slugify
from django.utils.translation import ugettext as _

from vega_admin.forms import ListViewSearchForm


class VegaFormKwargsMixin:  # pylint: disable=too-few-public-methods
    """Adds form kwargs."""

    def get_form_kwargs(self):
        """Add kwargs to the form."""
        kwargs = super().get_form_kwargs()
        kwargs["request"] = self.request
        return kwargs


# pylint: disable=too-few-public-methods
class VegaFormMixin(VegaFormKwargsMixin):
    """Adds some nice stuff to formviews used in create/update/delete views."""


class VegaOrderedQuerysetMixin:
    """Optionally ensures querysets are ordered."""

    order_by = None

    def get_order_by(self):
        """Get the field to use when ordering."""
        return self.order_by or settings.VEGA_ORDERING_FIELD

    def get_queryset(self):
        """Get the queryset."""
        queryset = super().get_queryset()
        if not queryset.ordered and settings.VEGA_FORCE_ORDERING:
            order_by = self.get_order_by()
            queryset = queryset.order_by(*order_by)

        return queryset


class ListViewSearchMixin:
    """Adds search to listview."""

    form_class = ListViewSearchForm
    search_fields: List[str] = []
    filter_class = None

    def get_queryset(self):
        """Get the queryset."""
        queryset = super().get_queryset()

        if self.filter_class:
            # pylint: disable=not-callable
            the_filter = self.filter_class(self.request.GET, queryset=queryset)
            queryset = the_filter.qs

        if self.request.GET.get("q"):
            form = self.form_class(self.request.GET)
            if form.is_valid() and self.search_fields:
                search_terms = [f"{x}__icontains" for x in self.search_fields]
                query = Q()
                for term in search_terms:
                    query.add(Q(**{term: form.cleaned_data["q"]}), Q.OR)
                queryset = queryset.filter(query)

        return queryset.distinct()

    def get_search_form_values(self):
        """Get search form values."""
        fields = []
        if self.filter_class:
            for field in self.filter_class.Meta.fields:
                fields.append(field)
        if self.search_fields:
            fields.append("q")
        # build the form values dict
        result = {}
        for item in fields:
            result[item] = self.request.GET.get(item)

        return result

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        if self.search_fields or self.filter_class:
            form = self.form_class(request=self.request)
            initial_values = self.get_search_form_values()
            form = self.form_class(initial=initial_values)
            context["vega_listview_search_form"] = form

        return context


class VerboseNameMixin:
    """Sets the Model verbose name in the context data."""

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["vega_verbose_name"] = self.model._meta.verbose_name
        context["vega_verbose_name_plural"] = self.model._meta.verbose_name_plural
        return context


class PageTitleMixin:
    """Sets the page title in the context data."""

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["vega_page_title"] = getattr(self, "page_title", None)
        return context


class CRUDURLsMixin:
    """Mixin that adds some CRUD helper urls."""

    cancel_url = "/"
    cancel_url_name = None
    delete_url = "/"
    delete_url_name = None
    read_url = "/"
    read_url_name = None
    list_url = "/"
    list_url_name = None
    create_url = "/"
    create_url_name = None
    update_url = "/"
    update_url_name = None

    def get_crud_url(  # pylint: disable=no-self-use,bad-continuation
        self, url: str, url_name: str, url_kwargs: dict = None
    ):
        """
        Return CRUD url.

        Attempt to use reverse_lazy to get the url, otherwise return what we
        hope are safe defaults

        :param url: the url
        :param url_name: the url name
        :param url_kwargs: the url kwargs

        :return: url
        """
        if url_name:
            if url_kwargs:
                return reverse_lazy(url_name, kwargs=url_kwargs)
            return reverse_lazy(url_name)
        return url

    def get_create_url(self):
        """
        Get the create url for the object in question.

        :return: url
        """
        return self.get_crud_url(url=self.create_url, url_name=self.create_url_name)

    def get_list_url(self):
        """
        Get the cancel url for the object in question.

        :return: url
        """
        return self.get_crud_url(url=self.list_url, url_name=self.list_url_name)

    def get_update_url(self):
        """
        Get the update url for the object in question.

        :return: url
        """
        return self.get_crud_url(
            url=self.update_url,
            url_name=self.update_url_name,
            url_kwargs={"pk": self.object.pk},
        )

    def get_read_url(self):
        """
        Get the read url for the object in question.

        :return: url
        """
        return self.get_crud_url(
            url=self.read_url,
            url_name=self.read_url_name,
            url_kwargs={"pk": self.object.pk},
        )

    def get_delete_url(self):
        """
        Get the delete url for the object in question.

        :return: url
        """
        return self.get_crud_url(
            url=self.delete_url,
            url_name=self.delete_url_name,
            url_kwargs={"pk": self.object.pk},
        )

    def get_cancel_url(self):
        """
        Get the cancel url for the object in question.

        :return: url
        """
        return self.get_crud_url(url=self.cancel_url, url_name=self.cancel_url_name)

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["vega_create_url"] = self.get_create_url()
        context["vega_list_url"] = self.get_list_url()
        context["vega_cancel_url"] = self.get_cancel_url()
        if hasattr(self, "object") and self.object is not None:
            context["vega_read_url"] = self.get_read_url()
            context["vega_delete_url"] = self.get_delete_url()
            context["vega_update_url"] = self.get_update_url()
        return context

    def get_form_kwargs(self):
        """Add kwargs to the form."""
        kwargs = super().get_form_kwargs()
        url_kwargs = {"cancel_url": self.get_list_url()}
        kwargs[settings.VEGA_MODELFORM_KWARG] = url_kwargs
        return kwargs


class ObjectTitleMixin:
    """Mixin for getting object title."""

    def get_title(self):
        """By default we just return the string representation of our object."""
        return str(self.object)

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["vega_object_title"] = self.get_title()
        return context


class DetailViewMixin:
    """Mixin for detail views."""

    fields = None

    def get_fields(self):
        """
        Get fields.

        We first default to using our 'fields' variable if available,
        otherwise we figure it out from our object.
        """
        if self.fields and isinstance(self.fields, list):
            return self.fields
        return [_.name for _ in self.object._meta.fields]

    def get_field_value(self, field):
        """Get the value of a field."""
        if field.is_relation:
            try:
                return str(getattr(self.object, field.name))
            except AttributeError:
                return None
        # pylint: disable=protected-access
        return self.object._get_FIELD_display(field)

    def get_object_data(self):
        """Return a dict of the data in the object."""
        result = {}
        fields_list = self.get_fields()
        for item in fields_list:
            try:
                field = self.object._meta.get_field(item)
            except FieldDoesNotExist:
                pass
            else:
                result[field.verbose_name] = self.get_field_value(field)

        return result

    def get_context_data(self, **kwargs):
        """Get context data."""
        context = super().get_context_data(**kwargs)
        context["vega_read_fields"] = self.get_fields()
        context["vega_object_data"] = self.get_object_data()
        return context


class DeleteViewMixin:
    """Mixin for delete views that adds in missing elements."""

    def delete(self, request, *args, **kwargs):
        """Delete method."""
        # Handle cases where you get ProtectedError
        try:
            return super().delete(request, *args, **kwargs)
        except ProtectedError:
            info = _(settings.VEGA_DELETE_PROTECTED_ERROR_TXT)
            messages.error(request, info, fail_silently=True)

            return redirect(self.get_delete_url())


class CRUDPathPatterMixin:
    """Add method to get CRUD path URL pattern"""

    @classmethod
    def get_crud_path_pattern(cls, crud_path: str, action: str):
        """Derive the CRUD path URL pattern."""
        return f"{crud_path}/{slugify(action, allow_unicode=True)}"


class SimpleURLPatternMixin(CRUDPathPatterMixin):
    """Very simple implements the derive_url_pattern method."""

    @classmethod
    def derive_url_pattern(cls, crud_path: str, action: str):
        """Derive the url pattern."""
        return f"{cls.get_crud_path_pattern(crud_path, action)}/"


class ObjectURLPatternMixin(CRUDPathPatterMixin):
    """Implements the derive_url_pattern method for single object views."""

    @classmethod
    def derive_url_pattern(cls, crud_path: str, action: str):
        """Derive the url pattern."""
        return f"{cls.get_crud_path_pattern(crud_path, action)}/<int:pk>/"


class TimeStampedModel(models.Model):
    """Adds timestamps to a model class."""

    created = models.DateTimeField(_("Created"), auto_now_add=True)
    modified = models.DateTimeField(_("Modified"), auto_now=True)

    class Meta:
        """Meta class."""

        abstract = True



"""Test vega_admin.contrib.users.forms module"""

from unittest.mock import patch

from django.test import TestCase, override_settings

from model_mommy import mommy

from vega_admin.contrib.users.forms import AddUserForm, EditUserForm, PasswordChangeForm


@override_settings(ROOT_URLCONF="vega_admin.contrib.users.urls", VEGA_TEMPLATE="basic")
class TestForms(TestCase):
    """
    Test class for vega_admin.contrib.users.forms
    """

    def test_adduserform(self):
        """
        Test AddUserForm

        """
        good_data = {
            "first_name": "mosh",
            "last_name": "pitt",
            "username": "moshthepitt",
            "email": "mosh@example.com",
            "password": "Miserable-Water-9",
        }

        form = AddUserForm(data=good_data)
        self.assertTrue(form.is_valid())
        user = form.save()
        self.assertEqual("mosh", user.first_name)
        self.assertEqual("pitt", user.last_name)
        self.assertEqual("moshthepitt", user.username)
        self.assertEqual("mosh@example.com", user.email)
        self.assertTrue(
            self.client.login(username="moshthepitt", password="Miserable-Water-9")
        )

        # good data no username
        good_data = {
            "first_name": "mosh",
            "last_name": "pitt",
            "email": "mosh2@example.com",
            "password": "Bean-Quarrel-0",
        }

        form = AddUserForm(data=good_data)
        self.assertTrue(form.is_valid())
        user = form.save()
        self.assertEqual("mosh", user.username)  # was generated

        # weak password
        bad_data = {
            "first_name": "mosh",
            "last_name": "pitt",
            "username": "moshthepitt2",
            "email": "mosh3@example.com",
            "password": "mosh@example.com",
        }
        form = AddUserForm(data=bad_data)
        self.assertFalse(form.is_valid())
        self.assertEqual(1, len(form.errors.keys()))
        self.assertEqual(
            "The password is too similar to the email address.",
            form.errors["password"][0],
        )

        # email not unique
        bad_data = {
            "first_name": "mosh",
            "last_name": "pitt",
            "username": "moshthepitt3",
            "email": "mosh@example.com",
            "password": "Every-Actor-1",
        }
        form = AddUserForm(data=bad_data)
        self.assertFalse(form.is_valid())
        self.assertEqual(1, len(form.errors.keys()))
        self.assertEqual(
            "A user is already registered with this e-mail address.",
            form.errors["email"][0],
        )

        # missing email and username
        bad_data = {
            "first_name": "mosh",
            "last_name": "pitt",
            "password": "TestAddUserForm",
        }
        form = AddUserForm(data=bad_data)
        self.assertFalse(form.is_valid())
        self.assertEqual(1, len(form.errors.keys()))
        self.assertEqual(
            "You must provide one of email or username", form.errors["__all__"][0]
        )

    def test_edituserform(self):
        """
        Test EditUserForm

        """
        user = mommy.make("auth.User")

        good_data = {
            "first_name": "mosh",
            "last_name": "pitt",
            "username": "moshthepitt22",
            "email": "mosh22@example.com",
        }
        form = EditUserForm(instance=user, data=good_data)
        self.assertTrue(form.is_valid())
        user = form.save()
        self.assertEqual("mosh", user.first_name)
        self.assertEqual("pitt", user.last_name)
        self.assertEqual("moshthepitt22", user.username)
        self.assertEqual("mosh22@example.com", user.email)

    def test_password_change_form(self):
        """
        Test PasswordChangeForm
        """
        user = mommy.make("auth.User", username="softie")

        good_data = {
            "password1": "PasswordChangeForm",
            "password2": "PasswordChangeForm",
        }
        form = PasswordChangeForm(instance=user, data=good_data)
        self.assertTrue(form.is_valid())
        form.save()
        self.assertTrue(
            self.client.login(username="softie", password="PasswordChangeForm")
        )

        # weak password
        bad_data = {"password1": "123456789", "password2": "123456789"}
        form = PasswordChangeForm(instance=user, data=bad_data)
        self.assertFalse(form.is_valid())
        self.assertEqual(1, len(form.errors.keys()))
        self.assertEqual("This password is too common.", form.errors["password2"][0])
        self.assertEqual(
            "This password is entirely numeric.", form.errors["password2"][1]
        )

        # different passwords
        bad_data = {"password1": "PasswordChangeForm", "password2": "123456789"}
        form = PasswordChangeForm(instance=user, data=bad_data)
        self.assertFalse(form.is_valid())
        self.assertEqual(1, len(form.errors.keys()))
        # we have to replace ’ with '
        # django 3.0 uses ’ while django 2.2 uses '
        self.assertEqual(
            "The two password fields didn't match.",
            form.errors["password2"][0].replace("’", "'"),
        )

    @patch("vega_admin.contrib.users.forms.get_form_actions")
    def test_password_change_form_cancel_url(self, mock):
        """Test cancel url on password change form"""
        user = mommy.make("auth.User")

        good_data = {
            "password1": "PasswordChangeForm",
            "password2": "PasswordChangeForm",
        }
        form = PasswordChangeForm(instance=user, data=good_data)
        self.assertTrue(form.is_valid())
        form.save()

        mock.assert_called_with(cancel_url="/auth.user/list/")



"""Module for vega-admin test views."""
from typing import List, Union

from django.views.generic import TemplateView

from braces.views import LoginRequiredMixin, PermissionRequiredMixin
from django_filters import FilterSet

from vega_admin.mixins import SimpleURLPatternMixin
from vega_admin.views import (
    VegaCreateView,
    VegaCRUDView,
    VegaDeleteView,
    VegaDetailView,
    VegaListView,
    VegaUpdateView,
)

from .forms import (
    ArtistForm,
    CustomSearchForm,
    PlainArtistForm,
    SongForm,
    UpdateArtistForm,
)
from .models import Artist, Band, Song
from .tables import ArtistTable


class ArtistCreate(VegaCreateView):  # pylint: disable=too-many-ancestors
    """Artist CreateView class."""

    form_class = ArtistForm
    model = Artist

    def get_success_url(self):
        """Get success URL."""
        return "/edit/artists/create/"


class ArtistUpdate(VegaUpdateView):  # pylint: disable=too-many-ancestors
    """Artist UpdateView class."""

    form_class = ArtistForm
    model = Artist

    def get_success_url(self):
        """Get success URL."""
        return f"/edit/artists/edit/{self.object.id}"


class ArtistDelete(VegaDeleteView):  # pylint: disable=too-many-ancestors
    """Artist DeleteView class."""

    model = Artist

    def get_success_url(self):
        """Get success URL."""
        return "/edit/artists/create/"


class ArtistRead(VegaDetailView):  # pylint: disable=too-many-ancestors
    """Artist detail view."""

    model = Artist


class ArtistListView(VegaListView):  # pylint: disable=too-many-ancestors
    """Artist list view."""

    model = Artist
    table_class = ArtistTable
    search_fields = ["name"]


class SongCRUD(VegaCRUDView):
    """CRUD view for songs."""

    model = Song
    protected_actions: Union[None, List[str]] = None
    permissions_actions: Union[None, List[str]] = None
    list_fields = ["name", "artist"]
    read_fields = ["name", "artist"]
    table_attrs = {"class": "song-table"}
    table_actions = ["create", "update", "delete"]
    create_fields = ["name", "artist"]
    update_fields = ["name"]


class CustomSongCRUD(SongCRUD):
    """CRUD view for songs with login protection."""

    class CustomListView(ArtistListView):  # pylint: disable=too-many-ancestors
        """custom list view."""

    class FooView(SimpleURLPatternMixin, TemplateView):
        """random template view."""

        template_name = "artist_app/empty.html"

    protected_actions = ["create", "update", "delete", "template", "view"]
    permissions_actions: Union[None, List[str]] = None
    crud_path = "private-songs"
    view_classes = {"artists": CustomListView, "template": FooView}
    form_fields = ["name", "artist"]


class PermsSongCRUD(CustomSongCRUD):
    """CRUD view for songs with permissions protection."""

    protected_actions = ["create", "update", "delete", "artists", "list", "view"]
    permissions_actions = ["create", "update", "delete", "artists", "view"]
    crud_path = "hidden-songs"
    form_class = SongForm


class ArtistCRUD(VegaCRUDView):
    """CRUD view for artists."""

    model = Artist
    protected_actions = None
    permissions_actions = None


class CustomDefaultActions(ArtistCRUD):
    """CRUD view with custom default actions."""

    # pylint: disable=too-many-ancestors
    class CustomCreateView(ArtistCreate):
        """custom Create view."""

    class CustomUpdateView(ArtistUpdate):
        """custom Update view."""

    class CustomListView(ArtistListView):
        """custom list view."""

    class CustomDeleteView(ArtistDelete):
        """custom Delete view."""

    class CustomReadView(ArtistRead):
        """custom Read view."""

    view_classes = {
        "list": CustomListView,
        "view": CustomReadView,
        "update": CustomUpdateView,
        "create": CustomCreateView,
        "delete": CustomDeleteView,
    }
    crud_path = "custom-default-actions"


class BrokenCRUD(VegaCRUDView):
    """CRUD view with broken urls."""

    # pylint: disable=too-many-ancestors
    class BrokenListView(LoginRequiredMixin, VegaListView):
        """View that is broken."""

        model = Artist

    model = Artist
    actions = ["break"]
    permissions_actions = actions
    view_classes = {"break": BrokenListView}
    crud_path = "broken"


class Artist42CRUD(VegaCRUDView):
    """CRUD View that sets permission required for custom view class and action."""

    # pylint: disable=too-many-ancestors
    class CustListView(PermissionRequiredMixin, VegaListView):
        """Custom list view that has PermissionRequiredMixin."""

        model = Artist

    class FooView(SimpleURLPatternMixin, TemplateView):
        """random template view."""

        template_name = "artist_app/empty.html"

    model = Artist
    actions = ["list", "other"]
    permissions_actions = actions
    view_classes = {"list": CustListView, "other": FooView}
    crud_path = "42"


class CustomArtistCRUD(VegaCRUDView):
    """CRUD view for artists with custom forms and table."""

    model = Artist
    protected_actions = None
    permissions_actions = None
    crud_path = "custom-artist"
    form_class = PlainArtistForm
    create_form_class = ArtistForm
    update_form_class = UpdateArtistForm
    table_class = ArtistTable
    search_fields = ["name"]
    search_form_class = CustomSearchForm
    paginate_by = 10


class FilterSongCRUD(VegaCRUDView):
    """CRUD view for songs with filtering."""

    model = Song
    actions = ["list"]
    filter_fields = ["name", "artist"]
    crud_path = "filters"
    search_form_class = None


class Filter2SongCRUD(VegaCRUDView):
    """CRUD view for songs with filtering."""

    class SongFilter(FilterSet):
        """Song filter class."""

        class Meta:
            """Meta options."""

            model = Song
            fields = ["artist"]

    model = Song
    actions = ["list"]
    search_fields = ["artist__name"]
    filter_class = SongFilter
    crud_path = "filters2"
    search_form_class = None


class PlainFormCRUD(VegaCRUDView):
    """Vega CRUD view created with plain form."""

    model = Artist
    permissions_actions = None
    form_class = PlainArtistForm
    actions = ["create", "update", "list"]
    crud_path = "plain-form"


class BandCRUD(VegaCRUDView):
    """CRUD view for songs."""

    model = Band
    protected_actions: Union[None, List[str]] = None
    permissions_actions: Union[None, List[str]] = None



"""
Apps module for django-vega-admin
"""
from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class VegaAdminConfig(AppConfig):
    """
    Apps config class
    """

    name = "vega_admin"
    app_label = "vega_admin"
    verbose_name = _("Vega Admin")

    def ready(self):
        """
        Do stuff when the app is ready
        """
        # set up app settings
        # pylint: disable=import-outside-toplevel
        from django.conf import settings
        import vega_admin.settings as defaults

        for name in dir(defaults):
            if name.isupper() and not hasattr(settings, name):
                setattr(settings, name, getattr(defaults, name))



"""Main init file for vega_admin."""
VERSION = (0, 0, 17)
__version__ = ".".join(str(v) for v in VERSION)
# pylint: disable=invalid-name
default_app_config = "vega_admin.apps.VegaAdminConfig"  # noqa



"""vega-admin forms module."""
from typing import Any, Dict, List, Optional, Tuple, Union

from django import forms
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import DateField, DateTimeField, Model, TimeField
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch
from django.utils.html import format_html
from django.utils.module_loading import import_string
from django.utils.translation import ugettext as _

import django_tables2 as tables
from django_filters import FilterSet

from vega_admin.crispy_utils import get_form_helper_class, get_layout
from vega_admin.mixins import VegaFormMixin


def get_datefields(model: Model) -> List[str]:
    """
    Get the date fields from a model.

    :param model: the model class
    :return: list of datefield names
    """
    return [
        _.name
        for _ in model._meta.concrete_fields
        if isinstance(_, DateField) and not isinstance(_, DateTimeField)
    ]


def get_datetimefields(model: Model) -> List[str]:
    """
    Get the datetime fields from a model.

    :param model: the model class
    :return: list of datetimefield names
    """
    return [_.name for _ in model._meta.concrete_fields if isinstance(_, DateTimeField)]


def get_timefields(model: Model) -> List[str]:
    """
    Get the time fields from a model.

    :param model: the model class
    :return: list of timefield names
    """
    return [_.name for _ in model._meta.concrete_fields if isinstance(_, TimeField)]


def get_modelform(model: Model, fields: list = None, extra_fields: list = None):
    """
    Get the ModelForm for the provided model.

    :param model: the model class
    :param fields: list of the fields that you want included in the form
    :param extra_fields: extra fields that you want included in the form
    :return: model form

    extra_fields needs to be a list of tuples, like so:

    extra_fields = [("q", forms.CharField(
            label=_(settings.VEGA_LISTVIEW_SEARCH_QUERY_TXT),
            required=False,))]
    """
    # this is going to be our custom init method
    def _constructor(self, *args, **kwargs):
        self.request = kwargs.pop("request", None)
        self.vega_extra_kwargs = kwargs.pop(settings.VEGA_MODELFORM_KWARG, dict())
        super(modelform_class, self).__init__(*args, **kwargs)
        # add crispy forms FormHelper
        self.helper = get_form_helper_class(
            form_tag=True,
            form_method="POST",
            render_required_fields=True,
            form_show_labels=True,
            html5_required=True,
            include_media=True,
        )
        self.helper.form_id = f"{self.model._meta.model_name}-form"
        self.helper.layout = get_layout(
            self.fields.keys(),
            with_actions=True,
            cancel_url=self.vega_extra_kwargs.get("cancel_url", "/"),
        )

    if fields is None:
        fields = [field.name for field in model._meta.get_fields() if field.editable]

    widgets = {}
    # set the widgets for all date input fields
    for datefield in get_datefields(model):
        widgets[datefield] = import_string(settings.VEGA_DATE_WIDGET)

    # set the widgets for all datetime input fields
    for datetimefield in get_datetimefields(model):
        widgets[datetimefield] = import_string(settings.VEGA_DATETIME_WIDGET)

    # set the widgets for all time input fields
    for timefield in get_timefields(model):
        widgets[timefield] = import_string(settings.VEGA_TIME_WIDGET)

    meta_class_options = {"model": model, "fields": fields}

    if widgets:
        meta_class_options["widgets"] = widgets

    # the Meta class
    meta_class = type("Meta", (), meta_class_options)

    # the attributes of our new modelform
    options = {"model": model, "__init__": _constructor, "Meta": meta_class}

    # add extra fields
    if extra_fields:
        for extra_field in extra_fields:
            options[extra_field[0]] = extra_field[1]

    # create the modelform dynamically using type
    modelform_class = type(
        f"{model.__name__.title()}{settings.VEGA_FORM_LABEL}",
        (VegaFormMixin, forms.ModelForm),
        options,
    )

    return modelform_class


def get_listview_form(model: Model, fields: List[str], include_search: bool = True):
    """
    Get a search and filter form for use in ListViews.

    This is essentially a model form with an additional field named `q`.

    :param model: the model class
    :param fields: list of the fields that you want included in the form
    :return: model form
    """
    search_field: Optional[Tuple[str, Any]] = None
    if include_search:
        search_field = (
            "q",
            forms.CharField(
                label=_(settings.VEGA_LISTVIEW_SEARCH_QUERY_TXT), required=False
            ),
        )

    extra_fields: Optional[List[Tuple[str, Any]]] = None
    if search_field:
        extra_fields = [search_field]

    return get_modelform(model=model, fields=fields, extra_fields=extra_fields)


def get_table(  # pylint: disable=bad-continuation
    model: Model,
    fields: Optional[List[str]] = None,
    actions: Optional[List[str]] = None,
    attrs: Optional[dict] = None,
):
    """
    Get the Table Class for the provided model.

    :param model: the model class
    :param fields: list of the fields that you want included in the table
    :param actions: list of tuples representing actions and action url names
    :param options: dict representing kwargs/options to pass to the table
    :return: table
    """
    # the Meta class
    meta_options = {"model": model, "empty_text": _(settings.VEGA_NOTHING_TO_SHOW)}

    if fields:
        all_fields = [_.name for _ in model._meta.concrete_fields]
        exclude_fields = [_ for _ in all_fields if _ not in fields]
        # get sequence
        sequence_list = fields
        if "..." not in sequence_list:
            sequence_list.append("...")
        # set meta options
        meta_options["exclude"] = exclude_fields
        meta_options["sequence"] = tuple(sequence_list)

    if attrs:
        meta_options["attrs"] = attrs

    meta_class = type("Meta", (), meta_options)

    # the attributes of our new table class
    options: Dict[Any, Any] = {"Meta": meta_class}

    if isinstance(actions, list):
        # pylint: disable=unused-argument
        def render_actions_fn(self, *args, **kwargs):
            """Render the actions column."""
            record = kwargs["record"]
            actions_links = []
            for item in self.actions_list:
                try:
                    url = reverse(item[1])
                except NoReverseMatch:
                    url = reverse(item[1], args=[record.pk])
                name = item[0]
                actions_links.append(f"<a href='{url}' class='vega-action'>{name}</a>")
            actions_links_html = settings.VEGA_ACTION_LINK_SEPARATOR.join(actions_links)
            return format_html(actions_links_html)

        options["actions_list"] = actions
        options["action"] = tables.Column(
            verbose_name=_(settings.VEGA_ACTION_COLUMN_NAME),
            accessor=settings.VEGA_ACTION_COLUMN_ACCESSOR_FIELD,
            orderable=False,
        )
        options["render_action"] = render_actions_fn

    # create the table dynamically using type
    table_class = type(
        f"{model.__name__.title()}{settings.VEGA_TABLE_LABEL}", (tables.Table,), options
    )

    return table_class


def get_filterclass(model: Model, fields: list = None):
    """
    Get the Filter Class for the provided model.

    :param model: the model class
    :param fields: list of the fields that you want included in the table
    :return: filter class
    """
    # the Meta class
    meta_options = {"model": model, "fields": fields}
    meta_class = type("Meta", (), meta_options)

    # the attributes of our new table class
    options = {"Meta": meta_class}

    # create the filter_class dynamically using type
    filter_class = type(
        f"{model.__name__.title()}{settings.VEGA_FILTER_LABEL}", (FilterSet,), options
    )

    return filter_class


def customize_modelform(form_class: Union[forms.Form, forms.ModelForm]):
    """
    Add custom keyword arguments to a provided form class.

    Adds the custom kwargs required for vega_admin, if they are missing.

    Arguments:
        form_class {Union[Form, ModelForm]} -- the form class

    Returns:
        {Union[Form, ModelForm]} -- the customized form class

    """
    # the try here would fail with a TypeError if settings.VEGA_MODELFORM_KWARG
    # is not a valid kwarg
    try:
        form_class(**{settings.VEGA_MODELFORM_KWARG: dict()})
    except ObjectDoesNotExist:
        # we don't care about these exceptions right now because forms that have
        # an instance are likely to end up here as we are not supplying that
        # instance in the try statements
        pass
    except TypeError:
        # pylint: disable=missing-class-docstring,too-few-public-methods,inherit-non-class
        class VegaCustomFormClass(form_class):  # type: ignore
            def __init__(self, *args, **kwargs):
                self.request = kwargs.pop("request", None)
                self.vega_extra_kwargs = kwargs.pop(
                    settings.VEGA_MODELFORM_KWARG, dict()
                )
                super().__init__(*args, **kwargs)

        return VegaCustomFormClass

    return form_class



"""Module for vega-admin test urls."""
from django.urls import path

from . import views

# pylint: disable=invalid-name
artist_crud_patterns = views.ArtistCRUD().url_patterns()
custom_artist_crud_patterns = views.CustomArtistCRUD().url_patterns()
song_crud_patterns = views.SongCRUD().url_patterns()
protected_crud_song_patterns = views.CustomSongCRUD().url_patterns()
perms_crud_song_patterns = views.PermsSongCRUD().url_patterns()
custom_default_patterns = views.CustomDefaultActions().url_patterns()
patterns_42 = views.Artist42CRUD().url_patterns()
plainform_patterns = views.PlainFormCRUD().url_patterns()


urlpatterns = (
    [
        path("list/artists/", views.ArtistListView.as_view()),
        path("edit/artists/create/", views.ArtistCreate.as_view()),
        path("view/artists/view/<int:pk>", views.ArtistRead.as_view()),
        path("edit/artists/edit/<int:pk>", views.ArtistUpdate.as_view()),
        path("edit/artists/delete/<int:pk>", views.ArtistDelete.as_view()),
    ]
    + artist_crud_patterns
    + song_crud_patterns
    + custom_artist_crud_patterns
    + protected_crud_song_patterns
    + custom_default_patterns
    + perms_crud_song_patterns
    + patterns_42
    + plainform_patterns
)



"""Test vega_admin.contrib.users.forms module"""

from django.test import TestCase, override_settings

from model_mommy import mommy

from vega_admin.contrib.users.forms import PasswordChangeForm
from vega_admin.contrib.users.views import ChangePassword
from vega_admin.views import VegaUpdateView


@override_settings(ROOT_URLCONF="vega_admin.contrib.users.urls", VEGA_TEMPLATE="basic")
class TestViews(TestCase):
    """
    Test class for vega_admin.contrib.users.views
    """

    def test_changepassword(self):
        """
        Test ChangePassword

        """
        user = mommy.make("auth.User", username="TestChangePasswordView")
        data = {
            "password1": "Extension-I-School-5",
            "password2": "Extension-I-School-5",
        }

        res = self.client.get(f"/auth.user/change-password/{user.id}/")
        self.assertEqual(res.status_code, 200)
        self.assertIsInstance(res.context["form"], PasswordChangeForm)
        self.assertIsInstance(res.context["view"], ChangePassword)
        self.assertIsInstance(res.context["view"], VegaUpdateView)
        self.assertTemplateUsed(res, "vega_admin/basic/update.html")
        res = self.client.post(f"/auth.user/change-password/{user.id}/", data)
        self.assertEqual(res.status_code, 302)
        self.assertRedirects(res, "/auth.user/list/")
        user.refresh_from_db()
        self.assertTrue(
            self.client.login(
                username="TestChangePasswordView", password="Extension-I-School-5"
            )
        )

    @override_settings(VEGA_FORCE_ORDERING=True)
    def test_list_view_ordering(self):
        """
        Test VEGA_FORCE_ORDERING=True
        """
        mommy.make("auth.User", _quantity=7)
        res = self.client.get("/auth.user/list/")
        self.assertTrue(res.context["object_list"].ordered)
        self.assertEqual("-last_login", res.context["object_list"].query.order_by[0])
        self.assertEqual("first_name", res.context["object_list"].query.order_by[1])

    @override_settings(VEGA_FORCE_ORDERING=False)
    def test_list_view_ordering_off(self):
        """
        Test VEGA_FORCE_ORDERING=False
        """
        mommy.make("auth.User", _quantity=3)
        res = self.client.get("/auth.user/list/")
        self.assertFalse(res.context["object_list"].ordered)



"""
vega-admin settings module
"""
from django.conf import settings

# general
VEGA_CREATE_ACTION = "create"
VEGA_READ_ACTION = "view"
VEGA_UPDATE_ACTION = "update"
VEGA_LIST_ACTION = "list"
VEGA_DELETE_ACTION = "delete"
VEGA_DEFAULT_ACTIONS = [
    VEGA_CREATE_ACTION,
    VEGA_READ_ACTION,
    VEGA_UPDATE_ACTION,
    VEGA_LIST_ACTION,
    VEGA_DELETE_ACTION,
]
VEGA_TEMPLATE = "basic"
# ensures that listview queries are ordered
VEGA_FORCE_ORDERING = True
VEGA_ORDERING_FIELD = ["-pk"]

# model forms
VEGA_MODELFORM_KWARG = "vega_extra_kwargs"

# crispy forms
VEGA_CRISPY_TEMPLATE_PACK = getattr(settings, "CRISPY_TEMPLATE_PACK", "bootstrap3")

# strings
VEGA_FORM_VALID_CREATE_TXT = "Created successfully!"
VEGA_FORM_VALID_UPDATE_TXT = "Updated successfully!"
VEGA_FORM_VALID_DELETE_TXT = "Deleted successfully!"
VEGA_FORM_INVALID_TXT = "Please correct the errors on the form."
VEGA_DELETE_PROTECTED_ERROR_TXT = (
    "You cannot delete this item, it is referenced by other items."
)
VEGA_PERMREQUIRED_NOT_SET_TXT = "PermissionRequiredMixin not set for"
VEGA_LISTVIEW_SEARCH_TXT = "Search"
VEGA_LISTVIEW_SEARCH_QUERY_TXT = "Search Query"
VEGA_NOTHING_TO_SHOW = "Nothing to show"
VEGA_CANCEL_TEXT = "Cancel"
VEGA_SUBMIT_TEXT = "Submit"
VEGA_TABLE_LABEL = "Table"
VEGA_FORM_LABEL = "Form"
VEGA_VIEW_LABEL = "View"
VEGA_FILTER_LABEL = "Filter"
VEGA_PROTECTED_LABEL = "Protected"
VEGA_ACTION_COLUMN_NAME = ""
VEGA_ACTION_COLUMN_ACCESSOR_FIELD = "pk"
VEGA_ACTION_LINK_SEPARATOR = " | "
VEGA_CHANGE_PASSWORD_LABEL = "change password"

# exceptions
VEGA_INVALID_ACTION = "Invalid Action"

# widgets
VEGA_DATE_WIDGET = "vega_admin.widgets.VegaDateWidget"
VEGA_DATETIME_WIDGET = "vega_admin.widgets.VegaDateTimeWidget"
VEGA_TIME_WIDGET = "vega_admin.widgets.VegaTimeWidget"

# contrib
# users
VEGA_USERNAME_HELP_TEXT = (
    "Optional. 150 characters or fewer. Letters, digits and @/./+/-/_ only."
)
VEGA_OPTIONAL_TXT = "Optional."
VEGA_EMAIL_OR_USERNAME_REQUIRED_TXT = "You must provide one of email or username"



"""Module for vega-admin test models."""
from django.db import models


class Artist(models.Model):
    """Artist Model class."""

    name = models.CharField(max_length=100)

    class Meta:
        """Meta class def."""

        ordering = ["name"]
        verbose_name = "professional artist"
        verbose_name_plural = "professional artists"

    def __str__(self):
        """Unicode representation of Artist."""
        return self.name


class Song(models.Model):
    """Model definition for Song."""

    SINGLE = "1"
    COLLABO = "2"
    SKIT = "3"

    SONG_TYPES = ((SINGLE, "Single"), (COLLABO, "Collaboration"), (SKIT, "Skit"))

    artist = models.ForeignKey(Artist, on_delete=models.PROTECT)
    name = models.CharField(max_length=100)
    song_type = models.CharField(
        "Type", max_length=1, choices=SONG_TYPES, default=SINGLE
    )
    release_date = models.DateField("Release Date")
    release_time = models.TimeField("Release Time")
    recording_time = models.DateTimeField("Recording Time", auto_now_add=False)

    class Meta:
        """Meta definition for Song."""

        verbose_name = "Song"
        verbose_name_plural = "Songs"
        ordering = ["name"]

    def __str__(self):
        """Unicode representation of Song."""
        return self.name


class Band(models.Model):
    """Band Model class."""

    created = models.DateTimeField(auto_now_add=True)
    name = models.CharField(max_length=100)

    class Meta:
        """Meta class def."""

        ordering = ["created"]
        verbose_name = "band"
        verbose_name_plural = "bands"

    def __str__(self):
        """Unicode representation of Band."""
        return self.name



"""AppConfig module for Vega Admin users app"""
from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class UsersConfig(AppConfig):
    """App config class"""

    name = "vega_admin.contrib.users"
    app_label = "vega_admin"
    verbose_name = _("Vega Admin Users")



"""Setup.py for django-vega-admin."""
import os
import sys

from setuptools import find_packages, setup

import vega_admin

with open(os.path.join(os.path.dirname(__file__), "README.md")) as readme:
    README = readme.read()

if sys.argv[-1] == "publish":
    if os.system("pip freeze | grep twine"):
        print("twine not installed.\nUse `pip install twine`.\nExiting.")
        sys.exit()
    os.system("python setup.py sdist bdist_wheel")
    os.system("twine upload dist/* --skip-existing")
    print("You probably want to also tag the version now:")
    print(
        f"  git tag -a {vega_admin.__version__} -m 'version {vega_admin.__version__}'"
    )
    print("  git push --tags")
    sys.exit()

setup(
    name="django-vega-admin",
    version=vega_admin.__version__,
    description="Simple and fast automated CRUD for any Django model",
    long_description=README,
    long_description_content_type="text/markdown",
    license="MIT",
    author="Kelvin Jayanoris",
    author_email="kelvin@jayanoris.com",
    url="https://github.com/moshthepitt/django-vega-admin",
    packages=find_packages(exclude=["docs", "tests"]),
    install_requires=[
        "Django >=2.2",
        "django-crispy-forms",
        "django-braces",
        "django-filter",
        "django-tables2",
        "tablib",
        "pyyaml>=4.2b1",  # fixes security vulnerability
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
    ],
    include_package_data=True,
)



"""Forms module for users"""
from django import forms
from django.conf import settings
from django.contrib.auth import password_validation
from django.contrib.auth.forms import AdminPasswordChangeForm
from django.contrib.auth.models import User
from django.urls import reverse_lazy
from django.utils.translation import ugettext as _

from crispy_forms.bootstrap import Field
from crispy_forms.layout import Layout

from vega_admin.crispy_utils import get_form_actions, get_form_helper_class

try:
    # pylint: disable=import-error
    from allauth.account import app_settings
    from allauth.account.adapter import get_adapter
except ModuleNotFoundError:
    UNIQUE_EMAIL = False
else:
    UNIQUE_EMAIL = app_settings.UNIQUE_EMAIL  # pylint: disable=no-member


def validate_unique_email(value):
    """Validate unique email when editting users"""
    try:
        return get_adapter().validate_unique_email(email=value)
    except NameError:
        return value


def get_formhelper():
    """
    Get form helper class.
    This is simply for convenience and to avoid pylint warning about
    duplicate code.
    """
    return get_form_helper_class(
        form_method="POST",
        form_tag=True,
        form_show_labels=True,
        include_media=True,
        render_required_fields=True,
        html5_required=True,
    )


class UserFormMixin:  # pylint: disable=too-few-public-methods
    """User forms mixin"""

    def clean_email(self):
        """clean email address"""
        value = self.cleaned_data["email"]
        try:
            value = get_adapter().clean_email(value)
        except NameError:
            pass
        if value and UNIQUE_EMAIL:
            value = validate_unique_email(value)
        return value


class AddUserForm(UserFormMixin, forms.ModelForm):
    """
    Form class to add users
    """

    password = forms.CharField(
        label=_("Password"),
        strip=True,
        help_text=password_validation.password_validators_help_text_html(),
    )

    class Meta:
        model = User
        fields = [
            "first_name",
            "last_name",
            "username",
            "email",
            "password",
            "is_active",
        ]

    def __init__(self, *args, **kwargs):
        """init method for add user form"""
        self.request = kwargs.pop("request", None)
        self.vega_extra_kwargs = kwargs.pop("vega_extra_kwargs", dict())
        super().__init__(*args, **kwargs)
        self.fields["first_name"].required = True
        self.fields["last_name"].required = True
        self.fields["username"].required = False
        self.fields["username"].help_text = _(settings.VEGA_USERNAME_HELP_TEXT)
        self.fields["email"].help_text = _(settings.VEGA_OPTIONAL_TXT)
        self.helper = get_formhelper()
        self.helper.form_id = "add-user-form"
        self.helper.layout = Layout(
            Field("first_name"),
            Field("last_name"),
            Field("username"),
            Field("email"),
            Field("password", autocomplete="off"),
            Field("is_active"),
            get_form_actions(cancel_url=self.vega_extra_kwargs.get("cancel_url", "/")),
        )

    def clean_password(self):
        """clean password field"""
        data = self.cleaned_data["password"]
        user = User()
        if self.cleaned_data.get("first_name"):
            user.first_name = self.cleaned_data["first_name"]
        if self.cleaned_data.get("last_name"):
            user.last_name = self.cleaned_data["last_name"]
        if self.cleaned_data.get("email"):
            user.email = self.cleaned_data["email"]
        password_validation.validate_password(password=data, user=user)
        return data

    def validate_unique_email(self, value):  # pylint: disable=no-self-use
        """validate unique email while adding users"""
        try:
            return get_adapter().validate_unique_email(email=value)
        except NameError:
            return value

    def clean(self):
        """General clean method"""
        cleaned_data = super().clean()
        username = cleaned_data.get("username")
        email = cleaned_data.get("email")
        if not email and not username:
            raise forms.ValidationError(_(settings.VEGA_EMAIL_OR_USERNAME_REQUIRED_TXT))

    def save(self, commit=True):
        """general form save method"""
        unique_username = self.cleaned_data.get("username")
        if not unique_username:
            try:
                unique_username = get_adapter().generate_unique_username(
                    [
                        self.cleaned_data.get("first_name"),
                        self.cleaned_data.get("last_name"),
                        self.cleaned_data.get("email"),
                    ]
                )
            except NameError:
                unique_username = self.cleaned_data.get("email")
        user_data = dict(
            username=unique_username,
            first_name=self.cleaned_data["first_name"],
            last_name=self.cleaned_data["last_name"],
            password=self.cleaned_data["password"],
        )
        if self.cleaned_data.get("email"):
            user_data["email"] = self.cleaned_data.get("email")
        user = User.objects.create_user(**user_data)
        return user


class EditUserForm(UserFormMixin, forms.ModelForm):
    """Edit User form"""

    class Meta:
        model = User
        fields = ["first_name", "last_name", "username", "email", "is_active"]

    def __init__(self, *args, **kwargs):
        """init method for edit user form"""
        self.request = kwargs.pop("request", None)
        self.vega_extra_kwargs = kwargs.pop("vega_extra_kwargs", dict())
        super().__init__(*args, **kwargs)
        self.fields["email"].help_text = _(settings.VEGA_OPTIONAL_TXT)
        self.helper = get_formhelper()
        self.helper.form_id = "edit-user-form"
        self.helper.layout = Layout(
            Field("first_name"),
            Field("last_name"),
            Field("username"),
            Field("email"),
            Field("is_active"),
            get_form_actions(cancel_url=self.vega_extra_kwargs.get("cancel_url", "/")),
        )


class PasswordChangeForm(AdminPasswordChangeForm):
    """Custom form for admins to change user passwords"""

    def __init__(self, *args, **kwargs):
        """init method for password change form"""
        self.instance = kwargs.pop("instance", None)
        self.request = kwargs.pop("request", None)
        self.vega_extra_kwargs = kwargs.pop("vega_extra_kwargs", dict())
        super().__init__(user=self.instance, *args, **kwargs)
        self.helper = get_formhelper()
        self.helper.form_id = "change-password-form"
        self.helper.layout = Layout(
            Field("password1"),
            Field("password2"),
            get_form_actions(cancel_url=reverse_lazy("auth.user-list")),
        )
